/***************************
**
**   Random Graph Generator
**
**
***************************/

/* if a level includes i nodes which is less than number of requested edges
then each of them */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#define MAXDEPTH 500
#define MAX_OUT_DEGREE  500
#define MAXINT          0x7FFFFFFF
#define MAX(x,y)        ((x == MAXINT || x > y) ? x : y)
#define MIN(x,y)        ((x == MAXINT || x > y) ? y : x)
#define GREATER(x,y)    ((x == MAXINT || x > y) ? TRUE : FALSE)
#define ABS(x)          (((x) < 0.0) ? (-1.0 * (x)) : (x))
#define ERROR_LIMIT_CCR(x)    (((x) <= 0.2) ? 0.5 : 0.1)
#define ERROR_LIMIT_RANGE(x)  (((x) <= 0.2) ? 0.5 : 0.2)    
#define ERR_LIMIT       0.00001
#define MAXTPE          64
#define MAXNODE         4096
#define TRUE    1
#define FALSE   0
#define FILE_NAME       "TEST_FILE"
#define FILE_ERR        "ERROR_FILE"
/* #define RAND_MAX 32767 */

typedef struct str_link { /* child/parent of a graph node */ 
    int    nid; 
    float   commcost;
    struct str_link *linkNext; 
} LINK; 

typedef struct str_node { /* a graph node */ 
    float        compcost; 
    float        tlevel, blevel;
    int        childnum;
    int        visit, visit_p, pnum;
    int        nid,i;
    LINK       *child, *parent;
    float        start, end;
    int        schPe, scheduled;
    int        pseudo_p, pseudo_c;
    struct str_node *schNext, *schLast;
    /*  int        level; */
} NODE; 

FILE  *out, *err;
NODE *graph;
NODE *nodep, *pnodep, *prev, *sch;
LINK *ptr, *parentptr, *childptr;

void  set_depth();
void  allocate_memory(); 
void  generate_nodes();
void  generate_edges(int outdegree);
void  initialize_graph();
void  print_graph();
void  compute_actual_values();
void  save_graph_in_a_file();
void  free_memory();

int  connection[MAXNODE][MAX_OUT_DEGREE+1];
float  comp_weights[MAXNODE][MAXTPE];

int  numnode,            /* graph node information */
     depth,              /* depth of the graph */
     sibling[MAXDEPTH],  /* number of nodes at each layer */
     out_degree,         /* average outdegree */
     numsample,          /* number of samples */ 
     numtpe;             /* number of processors */

float CCR, avg_comp, avg_comm, range, actual_range, actual_CCR, shape;
int   *ready, *ready2, *ready_b, *ready2_b, readynum, readynum2, readynum_b, 
  readynum2_b, childnode;


void main(int argc, char **argv)
{
  int numsample,sum, k, stime, item=0;
  long  r_seed;
  char  g[50];
  
  if (argc < 8) {
    fprintf(stderr, "usage: %s numnode CCR shape out_degree range numtpe  
             seed \n", argv[0]);
    exit(0);
  }
  
  sscanf(argv[1],"%d",&numnode);
  sscanf(argv[2],"%f",&CCR);
  sscanf(argv[3],"%f",&shape);
  sscanf(argv[4],"%d",&out_degree);
  sscanf(argv[5],"%f",&range);
  sscanf(argv[6],"%d",&numtpe);
  sscanf(argv[7],"%ld",&r_seed);
  /* sprintf(g,"%d %f %d %f %d %f\n", numnode, shape, 
                out_degree, CCR, numtpe, range);*/

  /* fixed parameters */  
  srand(r_seed);
  do
    avg_comp = (float) ceil(2 * 10.0 * (double) rand() / (double) RAND_MAX);
  while (avg_comp < ERR_LIMIT);
  avg_comm = avg_comp * CCR;
  
  #ifdef STR
  while (1) {
     item++;
     set_depth();
     allocate_memory();
     generate_nodes();
     generate_edges(out_degree); 
     initialize_graph();
     compute_actual_values();
     if (item > 50) printf("ccr=%f range=%f\n", actual_CCR, actual_range); 
     if ((((actual_CCR / CCR) >= (1.0 - ERROR_LIMIT_CCR(CCR))) &&
	  ((actual_CCR / CCR) <=  (1.0 + ERROR_LIMIT_CCR(CCR)))) 
	                   &&
	  (((actual_range / range) >= (1.0 - ERROR_LIMIT_RANGE(range))) &&
	   ((actual_range / range) <=  (1.0 + ERROR_LIMIT_CCR(range))))) {
            save_graph_in_a_file();
            exit(1);
     }
     free_memory();	 
     /* printf("not accepted");*/
     if  ((item%50)==0) {
       do
	 avg_comp = (float) ceil(2 * 10.0 * (double) rand() / (double) RAND_MAX);
       while (avg_comp < ERR_LIMIT);
       avg_comm = avg_comp * CCR;
       srand(time(NULL));
       printf("resetting");
     }
     if  (item==1000) {
         if ((err = fopen(FILE_ERR, "a")) == NULL) {
             fprintf(stderr, "Error in writing graph file\n");
             exit(0);
	 }
	 fprintf(err,"%s\n", g);
	 fclose(err);
     }
  }
  #endif
  
  #ifdef N_STR
  set_depth();
  allocate_memory();
  generate_nodes();
  generate_edges(out_degree); 
  initialize_graph();
  compute_actual_values();
  save_graph_in_a_file();
  free_memory();	 
  #endif
}

void set_depth() {
    do
      depth = (int) ceil(2.0 * sqrt(1.0 * numnode) / shape * (double) rand() /
		     (double) RAND_MAX);
    while  (depth <= 2);
}

void generate_nodes() {
    int i, k, cur_nodes, diff, temp_level;    
    sibling[0] = sibling[depth-1]=1; sibling[depth]=0;
    cur_nodes = 2; /* start node and exit node */

    for (i=1; i<depth-1; i++) {
      do
	sibling[i] = (int) ceil (2.0 * shape * sqrt(1.0 * numnode) * (double) rand()/
				 (double) RAND_MAX); 
      while  (sibling[i] < 2);
      cur_nodes+=sibling[i];
    }
    diff = cur_nodes - numnode;
    
    /* if there are more node than numnode */
    if  (diff > 0)
      while  (diff > 0) {
	do
	    /* if 5 layers : 0..4, then the candidate layers: 1..3 */
	    temp_level  = (int) ceil((double) (depth-2) * (double) rand()/
				     (double) RAND_MAX);
	    
	while (sibling[temp_level] < 2);
	sibling[temp_level]--;
	diff--;
      }

    /* if there are not enough number of nodes */
    else  if (diff < 0)  
      while  (diff < 0) {
	do
	  temp_level  = (int) ceil((double) (depth-2) * (double) rand()/
				 (double) RAND_MAX);
	while (temp_level == 0);
	sibling[temp_level]++;
	diff++;
      }
  }

  void allocate_memory() {
    if ((graph = (NODE*) malloc ((numnode+1)*sizeof(NODE))) == NULL) { 
      fprintf(stderr, "Error in allocating memory, %d nodes\n",numnode);
      exit(0);
    }    
    if ((ready = (int*) malloc (numnode*sizeof(int))) == NULL) { 
        fprintf(stderr, "Error in allocating memory, %d nodes\n",numnode);
        exit(0);
    }
    if ((ready2 = (int*) malloc (numnode*sizeof(int))) == NULL) { 
        fprintf(stderr, "Error in allocating memory, %d nodes\n",numnode);
        exit(0);
    } 
    if ((ready_b = (int*) malloc (numnode*sizeof(int))) == NULL) { 
        fprintf(stderr, "Error in allocating memory, %d nodes\n",numnode);
        exit(0);
    }
    if ((ready2_b = (int*) malloc (numnode*sizeof(int))) == NULL) { 
        fprintf(stderr, "Error in allocating memory, %d nodes\n",numnode);
        exit(0);
    } 
  }
    
  void initialize_graph() {
    int i, j,m, entry_id;
    int temp_out_degree, temp_parent;
    float tempcost, weight_base, weight_mean, weight_range;
    
    for (i = readynum = readynum2 = 0, readynum_b = readynum2_b = 0; 
	                                       i < numnode; i++) { 
      /* initialization */
      graph[i].nid = i;
      graph[i].schPe = -1;
      graph[i].pnum  = 0;
      graph[i].child = graph[i].parent = NULL;
      graph[i].start = graph[i].end = graph[i].tlevel = graph[i].blevel = 0.0;
      graph[i].visit = graph[i].scheduled = 0;
      graph[i].pseudo_p = graph[i].pseudo_c = -1;

      /* then generate random comm costs */
      for (j = 0; j < connection[i][0]; j++) { 
	graph[i].childnum=connection[i][0];
	if ((ptr = (LINK*) malloc (sizeof(LINK))) == NULL) { 
	  fprintf(stderr, "ERROR : memory for edge of node %d\n", i+1); 
	  exit(0); 
	} 
        ptr->nid = connection[i][j+1];
	do
	  ptr->commcost= (float) (2.0 * avg_comm * (double) rand() / 
				    (double) RAND_MAX);
	while (ptr->commcost < ERR_LIMIT);
	ptr->linkNext = graph[i].child; graph[i].child = ptr; 
      } 
      if (graph[i].child==NULL) {
	ready_b[readynum_b++] = i; 
	ready[readynum++] = i;
      }
    }
    
    /* Build up parent list for each node */ 
    for (i = 0; i < numnode; i++) {
        for (ptr=graph[i].child; ptr!=NULL; ptr=ptr->linkNext) { 
	  childnode = ptr->nid; 
	  if ((parentptr = (LINK*) malloc (sizeof(LINK))) == NULL) { 
	    fprintf(stderr, "ERROR : memory for parent node of %d\n", i+1); 
	    exit(0); 
	  } 
	  parentptr->nid = i; 
	  parentptr->commcost = ptr->commcost; 
	  parentptr->linkNext = graph[childnode].parent;
	  graph[childnode].parent = parentptr;
	  graph[childnode].pnum++;
	}
	if (graph[i].parent==NULL)  {
	  ready2_b[readynum2_b++] = i;
	  ready2[readynum2++] =i;
	}	
    }


    /* if node does not have a parent assign a random parent */
    for (i = 1; i < numnode; i++) {
      if  (graph[i].pnum  == 0) {
        m= entry_id = 0;
	do {
	  entry_id += sibling[m];
	  m++;
	} while (entry_id+sibling[m] <= i);
        m--;
        if (sibling[m] == 1)
	  temp_parent =  entry_id - 1;
	else
	  do
	    temp_parent  =  entry_id - (int) ceil(sibling[m] * 
				       (double) rand() / (double) RAND_MAX); 
	  while (temp_parent == entry_id);
        /* add the child connection info */
	if ((ptr = (LINK*) malloc (sizeof(LINK))) == NULL) { 
	  fprintf(stderr, "ERROR : memory for edge of node %d\n", i+1); 
	  exit(0); 
	} 
        ptr->nid = i;
	do
	  tempcost = ptr->commcost= (float) ceil(2.0 * avg_comm * 
						 (double) rand() / (double) RAND_MAX);
	while (tempcost < ERR_LIMIT);
	ptr->linkNext = graph[temp_parent].child;
	graph[temp_parent].child = ptr; 
	graph[temp_parent].childnum++;
	
        /* add parent connection info */
	if ((parentptr = (LINK*) malloc (sizeof(LINK))) == NULL) { 
	    fprintf(stderr, "ERROR : memory for parent node of %d\n", i+1); 
	    exit(0); 
	  } 
	parentptr->nid = temp_parent; 
	parentptr->commcost = tempcost;
	parentptr->linkNext = graph[i].parent;
	graph[i].parent = parentptr;
	graph[i].pnum++;
      }
    }

    /* random computation costs */
    for (m = 0; m < numnode; m++) {
        do
	  weight_mean = (float) (2.0 * avg_comp * 
				 (double) rand() / (double) RAND_MAX);
	while  (weight_mean < ERR_LIMIT);
        weight_range = weight_mean * range;
	for (j = 0; j <numtpe; j++) 
	    /* 0 yerine 1'den basliyor */ 
	  do
	    comp_weights[m][j] = (float) (weight_range *
					  (double) rand() / (double) RAND_MAX +
					  (weight_mean - weight_range / 2));
	  while  (comp_weights[m][j] < ERR_LIMIT);
    }
  }

  void generate_edges(int outdegree) {
      int  i, j, k, n, m, z, temp_child, duplicate, temp_outdegree;
      int node_id = m = 0;

    /* fixed number of outgoing edges from each node to the random nodes at 
       the next layer */

    /* since we assume there is one entry node this one must be connected 
       at least to the all nodes at the next level */
    connection[node_id][m++] = sibling[1];
    for (k=0; k<sibling[1]; k++) 
      connection[node_id][m++] = k + 1;
	
    for (node_id=1, i=1; i<depth-2; node_id+=sibling[i], i++) {

      /* printing graph 1  
	 for (z=node_id; z<sibling[i]+node_id; z++)
	    printf("%d  ", z);  
	    printf("\n"); */

      temp_outdegree = outdegree; 
      if (sibling[i+1] < outdegree)
	temp_outdegree = sibling[i+1];

      /* only one at level i; so connect it to all nodes at the next */
      if (sibling[i] == 1)  { 
	m=0;
	connection[node_id][m++] = sibling[i+1];
	for (k=0; k<sibling[i+1]; k++)
	  connection[node_id][m++] = node_id+sibling[i]+k;
      }
      /* if next level has one node then connect all of this level to it */
      else if (sibling[i+1]==1) {
	for (k=0; k<sibling[i]; k++) {
	  m=0;
	  connection[node_id+k][m++] = 1;
	  connection[node_id+k][m] = node_id+sibling[i];
	}
      }
      else {
	for (j=0; j<sibling[i]; j++) {
	  m=0;
	  connection[node_id+j][m++] = temp_outdegree;
	  for (k=0; k<temp_outdegree; k++) {        
	    do {	
	      duplicate = FALSE;
	      do
		temp_child  =  (int) ceil(sibling[i+1] * 
					  (double) rand() / (double) RAND_MAX); 
	      while (temp_child == 0);
	      for (n=1; n<m; n++) {
		 if (connection[node_id+j][n] == node_id+sibling[i]+
		     temp_child-1) {
		   duplicate = TRUE;
		   break;
		 }
	      }
	    }  while (duplicate == TRUE);
	    connection[node_id+j][m++] = node_id+sibling[i]+temp_child-1;
	  }
	}
      }
    }

    /* for the nodes at the level just before the end node */
    /* printing graph 2 
       for (z=node_id; z<sibling[i]+node_id; z++)
         printf("%d  ", z); 
	 printf("\n");  */
    for (; node_id < numnode-1; node_id++) {
      connection[node_id][0]=1;
      connection[node_id][1]=numnode-1;
    }
    connection[node_id][0] = 0;
  }

void compute_actual_values() {
  int j, k,  edge;
  float  sum_comm, sum_comp, mean;
  double variance;
  actual_range = actual_CCR = 0.0;
  for  (sum_comm=sum_comp=0, j=edge=0;  j<numnode; j++) {
      nodep = &(graph[j]);
      for (mean=0.0, k=0; k<numtpe; k++)
          mean += comp_weights[j][k];
      mean/= numtpe;
      sum_comp += mean;
      for (variance=0.0, k=0; k<numtpe; k++)
          variance += pow((comp_weights[j][k]-mean),2.0) / numtpe;
      actual_range += (float) sqrt(variance * 12.0/ pow((double) mean, 2));
      for (ptr=nodep->child; ptr!=NULL; ptr=ptr->linkNext) {
	  edge++;
          sum_comm += ptr->commcost;
      }
  } 
  actual_CCR = (sum_comm / edge) / (sum_comp / numnode);
  actual_range /= numnode;
  /* printf("range= %f ccr = %f\n", actual_range, actual_CCR);  */
}

void free_memory() {
  int i;
  for  (i=0; i<numnode; i++) {
    nodep=&(graph[i]);
    free(nodep->child);
    free(nodep->parent);
    free(nodep);
  }
  free(graph);
  free(ready);
  free(ready2);
  free(ready_b);
  free(ready2_b);
}

void  print_graph () {
   int i,j;
   printf("%d\n", numnode);
   for (i=0; i<numnode; i++) {
     nodep=&(graph[i]);
     printf("%d\t%d\t", graph[i].nid+1, graph[i].childnum);
     for (ptr=nodep->child; ptr!=NULL; ptr=ptr->linkNext) 
       printf("%d %5.2f\t", ptr->nid+1, ptr->commcost);
     printf("\n");
   }
   for (i=0; i<numnode; i++) {
     printf("%d\t", i+1);
     for (j=0; j<numtpe; j++)
       printf("%5.2f\t",comp_weights[i][j]);
     printf("\n");
   }
}

void  save_graph_in_a_file () {
   int i,j;

   if ((out = fopen(FILE_NAME, "w")) == NULL) {
     fprintf(stderr, "Error in writing graph file\n");
     exit(0);
   }
   fprintf(out,"%d %f %f\n", numnode, actual_CCR, actual_range);
   for (i=0; i<numnode; i++) {
     nodep=&(graph[i]);
     fprintf(out, "%d\t%d\t", graph[i].nid+1, graph[i].childnum);
     for (ptr=nodep->child; ptr!=NULL; ptr=ptr->linkNext) 
       fprintf(out,"%d %5.2f\t", ptr->nid+1, ptr->commcost);
     fprintf(out,"\n");
   }
   for (i=0; i<numnode; i++) {
     fprintf(out,"%d\t", i+1);
     for (j=0; j<numtpe; j++)
       fprintf(out,"%5.2f\t",comp_weights[i][j]);
     fprintf(out,"\n");
   }
   free(graph);
   fclose(out);
}





      
  

